/*********** util.c
		Shared functions for nlist and nlistd
			Written by Erich Menge & Gabor Paller
					**************/


#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <setjmp.h>
#include <time.h>
#include <sys/timeb.h>
#include <errno.h>
#include "config.h"
#include	<stdarg.h>
#if defined SOLARIS
#include	<sys/varargs.h>
#endif
#define		N_EXTERN extern
#include	"commonvars.h"

struct ListNode {
  char car;
  struct ListNode *cdr;
};

char *getline(FILE *fp)
{
	struct ListNode *head, *tail;
	char c, *ans;
	int len = 1;

	head = tail = (struct ListNode *)malloc(sizeof(struct ListNode));

	while(((c = fgetc(fp)) != EOF) && (c != '\n'))
	{
		tail->cdr = (struct ListNode *)malloc(sizeof(struct ListNode));
		tail = tail->cdr;
		tail->car = c;
		len++;
	}
	tail->cdr = NULL;
	if((c == EOF) && (head == tail))
	{
		ans = NULL;
	}
	else
	{
		ans = (char *)malloc(len * sizeof(char));

		tail = head;
		head = head->cdr; /* Get rid of the "sentinel" */
		free(tail);

		len = 0;
		while(head)
		{
			ans[len] = head->car;
			tail = head;
			head = head->cdr;
			free(tail);
			len++;
		}
		ans[len] = '\0';
	}

	return(ans);
}

/* Initializes the common variables to a safe value */
void init_commonvars() {
  int	i;
  
  config = NULL;
  addr = NULL;
  content_encoding = NULL;
  sender_full_address = NULL;
  closed = 0;
  subscribe = 1;
  sender_banner = 0;
  SubsConfirmation = 0;
  LogLevel = 1;
  LogEnable = 1;
  LogMailLoop = 0;
  AutoUnsubscribe = 0;
  strcpy( smtp_server,DEFAULT_SMTP_SERVER );
  strcpy( domain_name,"" );
  strcpy( content_type,"" );
  strcpy( boundary,"" );
  strcpy( list_address,"nlist" );
  strcpy( from_address,"nlist" );
  strcpy( nntp_spooldir,"" );
  for( i = 0 ; i < MAX_PASSPHRASE ; ++i )
  	passphrase[i][0] = 0;
}

/* Log function, writes its output into the log file */
void Log( int loglevel, int datestr, char format[], ... ) {
  va_list	ap;
  FILE	*logfile;
  LOCK_STRUCT	lt;
  int	i,err;
  struct timeb	ts;
  struct tm	*tp;
  char	timestr[32],*p,logbuf[1024];

  va_start( ap,format );
  if( loglevel < LogLevel ) {
    if( !LogEnable ) {
      if( loglevel < 1 )
        vfprintf( stderr,format,ap );
    } else {
      if( *list_name == 0 )
	    sprintf( LogFileName,"%s/general.log",PATH );
      else
	    sprintf( LogFileName,"%s/%s.log",PATH,list_name );
      if( ( logfile = fopen( LogFileName,"a" ) ) == NULL )
	goto l0;
      lt.l_type = F_WRLCK;
      lt.l_whence = 0;
      lt.l_start = 0;
      lt.l_len = 0;
/* We try ten times to lock the file */
      for( i = 0 ; i < 10 ; ++i ) {
        err = fcntl( fileno( logfile ), F_SETLK, &lt );
        if( err >= 0 )
	    break;
        if( ( err < 0 ) && ( errno == EAGAIN ) )
	    sleep( 1 );
      }
      if( err < 0 ) {
        fclose( logfile );
        fprintf( stderr, "Cannot lock log file %s for 10 seconds\n",
		    LogFileName );
        goto l0;
      }
      if( datestr ) {
        ftime( &ts );
        tp = localtime( &(ts.time) );
        strcpy( timestr,asctime( tp ) );
        timestr[ strlen( timestr ) - 1 ] = 0;
        fprintf( logfile,"%s ",timestr );
      }
      vsprintf( logbuf,format,ap );
      if( ( p = strchr( logbuf,'\n' ) ) != NULL )
          *p = 0;
      fprintf( logfile,"%s\n",logbuf );
      fclose( logfile );
    }
  }
l0:
  va_end( ap );
}

/* Error handling function. level=0 - warning,
   level=1 - fatal error */
void ehandle(int level, char format[], ...) {
  char	buf[N_BUFSIZE];
  va_list	ap;

  va_start( ap,format );
  switch (level) {
    case 1:
	vsprintf( buf,format,ap );
	Log( 0,1,"Error: %s",buf );
	longjmp( finish,GENERAL_ERROR );
	break;

    case 0:
	vsprintf( buf,format,ap );
	Log( 1,1,"Warning: %s",buf );
	break;
  }
  va_end( ap );
} 


/* Makes address substitution on a configuration parameter.
   Substitutes "sender" with act_from_address, "addressee" with
   act_to_address, "list" with list_address else leaves the
   input unchanged. */
char *subst_lname( char config_value[],
			    char *act_from_address,
			    char *act_to_address,
			    char *l_address ) {
  static char ret_buf[N_BUFSIZE];
  	
  if( strcasecmp( config_value,"sender" ) == 0 )
     strcpy( ret_buf,act_from_address );
  else
  if( strcasecmp( config_value,"addressee" ) == 0 )
      strcpy( ret_buf,act_to_address );
  else
  if( strcasecmp( config_value,"list" ) == 0 )
      strcpy( ret_buf,l_address );
  else
      strcpy( ret_buf,config_value );
  return ret_buf;
}


FILE *file_open( char fname[], char mode[] ) {
  FILE	*f;

  if( ( f = fopen( fname, mode ) ) ==NULL ) {
    if (errno == 2)
	ehandle(1, "File cannot be opened in mode %s: %s\n",
		mode,fname);
    else if (errno == 13)
	ehandle(1, "Permission to open %s denied\n",fname);
    else
	ehandle( 1,"Unknown error number %d\n", errno);
  }
  return f;
}

/* Processes a yes/no configuration option. Gets a pointer to the
   beginning of the parameter area, return 0 if the answer was no,
   1, if it was yes, -1, if it was neither */
int process_yes_no( char *c ) {
  char *p;
  if( ( p = strtok( c," \t\n" ) ) == NULL )
      return -1;
  if( strcasecmp( p,"yes" ) == 0 )
      return 1;
  if( strcasecmp( p,"no" ) == 0 )
      return 0;
  return -1;
}

/* Processes the config file of the list of the given name. Locks
   the config file but doesn't close it as it is the signal that
   one instance of the program is in the critical region.
   If the silent flag is nonzero, missing config file
   results in silent return, not error message */
void read_config( char listname[], int silent ) {
  char	fname[N_BUFSIZE];
  char	buf[N_BUFSIZE],*p,*pe;
  LOCK_STRUCT	lt;
  int	err,i,tc;
  int	passpctr;
  long	to;
  
  if( config != NULL )
      fclose( config );
  
  ConfigLockTimeout = 60L;
  SubsConfirmation = FALSE;
  SubsConfTimeout = 24;
  *archive = 0;
  closed = FALSE;
  subscribe = TRUE;
  sender_banner = FALSE;
  LogMailLoop = FALSE;
  AutoUnsubscribe = FALSE;
  strcpy( replyto,"" );
  strcpy( smtp_server,DEFAULT_SMTP_SERVER );
  strcpy( errors_to,"" );
  strcpy( to_address,"list" );
  strcpy( from_address,"sender" );
  strcpy( list_address,"" );
  strcpy( domain_name,"" );
  strcpy( list_title,"" );
  strcpy( nntp_newsgroup,"" );
  strcpy( nntp_server,DEFAULT_NNTP_SERVER );
  strcpy( nntp_list_address,"" );
  strcpy( nntp_from_address,"" );
  strcpy( nntp_replyto,"" );
  strcpy( filter_extensions,"" );
  strcpy( nntp_spooldir,"" );
  strcpy( monthindex_name,"" );
  nntp_port = 119;
  nntp_nobinary = FALSE;
  nntp_newstolist = FALSE;
  nntp_newschecktime = 1800;	/* News check time interval: 30 minutes */
  nntp_msgwindow = 60*60*24*2;	/* Message window: 2 days */
  nntp_msgexpiration = 60*60*24*30;	/* Message expiration: 30 days */
  nntp_newart_access = NEWART_NEWNEWS;
  nlist_umask = 022;
  garble_flag = FALSE;
  for( i = 0 ; i < MAX_PASSPHRASE ; ++i )
  	passphrase[i][0] = 0;
 
  sprintf( fname, "%s/%s.config", PATH, listname );
  if( ( config = fopen( fname,"r+" ) ) == NULL ) {
    if( silent )
	return;
    longjmp( finish,NO_LIST );
  }  
  while( fgets( buf,N_BUFSIZE-1,config ) != NULL ) {
    i = strlen( buf );
    if( i > 0 )
	--i;
    if( buf[i] == '\n' )
	buf[i] = 0;
    if( strncasecmp( "closed=",buf,7 ) == 0 ) {
      tc = process_yes_no( buf+7 );
      if( tc < 0 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
      closed = tc;
    } else
    if( strncasecmp( "subscribe=",buf,10 ) == 0 ) {
      tc = process_yes_no( buf+10 );
      if( tc < 0 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
      subscribe = tc;
    } else
    if( strncasecmp( "confirm_subscribe=",buf,18 ) == 0 ) {
      tc = process_yes_no( buf+18 );
      if( tc < 0 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
      SubsConfirmation = tc;
    } else
    if( strncasecmp( "confirm_subs_timeout=",buf,21 ) == 0 ) {
      if( sscanf( buf+21,"%u",&SubsConfTimeout ) < 1 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );

    } else
    if( strncasecmp( "senderbanner=",buf,13 ) == 0 ) {
      tc = process_yes_no( buf+13 );
      if( tc < 0 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
      sender_banner = tc;
    } else
    if( strncasecmp( "logging=",buf,8 ) == 0 ) {
      tc = process_yes_no( buf+8 );
      if( tc < 0 )
	ehandle( 1,"Error in %s config file: %s",listname,buf );
      LogEnable = tc; 
    } else
    if( strncasecmp( "garble=",buf,7 ) == 0 ) {
      tc = process_yes_no( buf+7 );
      if( tc < 0 )
	ehandle( 1,"Error in %s config file: %s",listname,buf );
      garble_flag = tc;       
    } else
    if( strncasecmp( "logloop=",buf,8 ) == 0 ) {
      tc = process_yes_no( buf+8 );
      if( tc < 0 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
      LogMailLoop = tc; 
    } else
    if( strncasecmp( "autounsubscribe=",buf,16 ) == 0 ) {
      tc = process_yes_no( buf+16 );
      if( tc < 0 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
      AutoUnsubscribe = tc; 
    } else
    if( strncasecmp( "replyto=",buf,8 ) == 0 ) {
      strcpy( replyto,buf+8 );
    } else
    if( strncasecmp( "errorsto=",buf,9 ) == 0 ) {
      strcpy( errors_to,buf+9 );
    } else
    if( strncasecmp( "to=",buf,3 ) == 0 ) {
      strcpy( to_address,buf+3 );
    } else
    if( strncasecmp( "from=",buf,5 ) == 0 ) {
      strcpy( from_address,buf+5 );
    } else
    if( strncasecmp( "archive=",buf,8 ) == 0 ) {
      strcpy( archive,buf+8 );
    } else
    if( strncasecmp( "list=",buf,5 ) == 0 ) {
     strcpy( list_address,buf+5 );
    } else
    if( strncasecmp( "smtp_server=",buf,12 ) == 0 ) {
      strcpy( smtp_server,buf+12 );
    } else
    if( strncasecmp( "domain_name=",buf,12 ) == 0 ) {
      strcpy( domain_name,buf+12 );
    } else
    if( strncasecmp( "title=",buf,6 ) == 0 ) {
      strcpy( list_title,buf+6 );
    } else
    if( strncasecmp( "filter_extensions=",buf,18 ) == 0 ) {
      strcpy( filter_extensions,buf+18 );
    } else
    if( strncasecmp( "timeout=",buf,8 ) == 0 ) {
      if( sscanf( buf+8,"%lu",&ConfigLockTimeout ) < 1 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
    } else
    if( strncasecmp( "loglevel=",buf,9 ) == 0 ) {
      if( sscanf( buf+9,"%u",&LogLevel ) < 1 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
    } else
    if( strncasecmp( "newsgroup=",buf,10 ) == 0 ) {
      strcpy( nntp_newsgroup,buf+10 );
    } else
    if( strncasecmp( "nntp_server=",buf,12 ) == 0 ) {
      strcpy( nntp_server,buf+12 );
    } else
    if( strncasecmp( "nntp_port=",buf,10 ) == 0 ) {
      if( sscanf( buf+10,"%u",&nntp_port ) < 1 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
    } else
    if( strncasecmp( "nntp_list=",buf,10 ) == 0 ) {
      strcpy( nntp_list_address,buf+10 );
    } else
    if( strncasecmp( "nntp_from=",buf,10 ) == 0 ) {
      strcpy( nntp_from_address,buf+10 );
    } else
    if( strncasecmp( "nntp_replyto=",buf,13 ) == 0 ) {
      strcpy( nntp_replyto,buf+13 );
    } else
    if( strncasecmp( "nntp_nobinary=",buf,14 ) == 0 ) {
      tc = process_yes_no( buf+14 );
      if( tc < 0 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
      nntp_nobinary = tc; 
    } else
    if( strncasecmp( "nntp_newstolist=",buf,16 ) == 0 ) {
      tc = process_yes_no( buf+16 );
      if( tc < 0 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
      nntp_newstolist = tc; 
    } else
    if( strncasecmp( "nntp_newschecktime=",buf,19 ) == 0 ) {
      if( sscanf( buf+19,"%u",&nntp_newschecktime ) < 1 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
    } else
    if( strncasecmp( "nntp_msgwindow=",buf,15 ) == 0 ) {
      if( sscanf( buf+15,"%u",&nntp_msgwindow ) < 1 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
    } else
    if( strncasecmp( "nntp_msgexpiration=",buf,19 ) == 0 ) {
      if( sscanf( buf+19,"%u",&nntp_msgexpiration ) < 1 )
        ehandle( 1,"Error in %s config file: %s",listname,buf );
    } else
    if( strncasecmp( "nntp_newart_access=",buf,19 ) == 0 ) {
      if( strncasecmp( "newnews",buf+19,7 ) == 0 )
      	nntp_newart_access = NEWART_NEWNEWS;
      else
      if( strncasecmp( "stat",buf+19,4 ) == 0 )
      	nntp_newart_access = NEWART_STAT;
      else
         ehandle( 1,"Error in %s config file: %s",listname,buf );
    } else
    if( strncasecmp( "nntp_spooldir=",buf,14 ) == 0 ) {
      strcpy( nntp_spooldir,buf+14 );
    } else
    if( strncasecmp( "umask=",buf,6 ) == 0 ) {
     if( sscanf( buf+6,"%o",&nlist_umask ) < 1 )
     	ehandle( 1,"Error in %s config file: %s",listname,buf ); 
    } else
    if( strncasecmp( "monthindex=",buf,11 ) == 0 ) {
      strcpy( monthindex_name,buf+11 );
    } else
    if( strncasecmp( "passphrase=",buf,11 ) == 0 ) {
      pe = buf+11;
      tc = 0;
      do {
        p = strtok( pe," \t" );
        pe = NULL;
        if( p != NULL ) {
          strcpy( passphrase[tc++],p );
          if( tc >= MAX_PASSPHRASE )
        	break;
        }
      } while( p != NULL );
    }
  }
  umask( nlist_umask );
  if( *list_address == 0 )
      ehandle( 1,"list address was not specified for list %s",listname );
  if( *nntp_list_address == 0 )
      strcpy( nntp_list_address,list_address );
  if( *nntp_from_address == 0 )
      strcpy( nntp_from_address,from_address );
  if( *nntp_replyto == 0 )
      strcpy( nntp_replyto,replyto );
/* We never write the config file neverthless we lock it for write.
   The reason is that this is the signal that one instance of the program
   is in the critical region where it may write common files. That's
   the reason why we close the config file only before quitting */
  lt.l_type = F_WRLCK;
  lt.l_whence = 0;
  lt.l_start = 0;
  lt.l_len = 0;
/* We try ten times to lock the file */
  for( to = 0 ; to < ConfigLockTimeout ; ++to ) {
    err = fcntl( fileno( config ), F_SETLK, &lt );
    if( err >= 0 )
      break;
    if( ( err < 0 ) && ( errno == EAGAIN ) )
        sleep( 1 );
  }
  if( err < 0 ) {    
    fclose( config );
    ehandle( 1, "Cannot lock config file %s for %lu seconds",fname,
		ConfigLockTimeout ); 
  }
}

/* Searches the letter for the line starting with the passphrase.
   Gets the letter file name, returns 0 if the passphrase was
   not found, 1 if it was found */
int find_passphrase( char letterfile[] ) {
  FILE	*letter;
  int	result,tc;
  char	fbuf[N_BUFSIZE];
  char	*p,*pe;
  
  letter = fopen( letterfile,"r" );
  if( letter == NULL )
  	return 0;
  fseek( letter, ContentPointer, SEEK_SET );
  while( fgets( fbuf,N_BUFSIZE-1,letter ) != NULL ) {
    pe = fbuf;
    result = 1;
    for( tc = 0 ; ( tc < MAX_PASSPHRASE ) && ( passphrase[tc][0] != 0 ) ; ++tc ) {
      p = strtok( pe," \t\r\n" );
      pe = NULL;
      if( ( p == NULL ) || 
          ( strcasecmp( p,passphrase[ tc ] ) != 0 ) ) {
        result = 0;
        break;
      }
    }
    if( result )
    	break; 
  }
  fclose( letter );
  return result;
}
